SystemOrganization addCategory: #'Graphics-Text'!

Rectangle subclass: #CharacterBlock
	instanceVariableNames: 'stringIndex text textLine'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CharacterBlock commentStamp: 'StephaneDucasse 5/18/2010 16:00' prior: 0!
I describe the location of one character displayed on the screen. My instances are used to return the results of methods:
	Paragraph characterBlockAtPoint: aPoint and
	Paragraph characterBlockForIndex: stringIndex.
Any recomposition or movement of a Paragraph can make the information I store stale.

text (Text): The text where my character is from
stringIndex (Integer): The index of my character in the text, starting from 1
textLine (TextLine): The displayed line my character is on
origin (Point): The top-left corner of the area allocated for displaying my
		character's glyph, in pixels, counting right then down from the
		top-left corner of the text display area, and starting from 0@0
corner (Point): The bottom-right corner of the area allocated for displaying my
		character's glyph, in pixels, counting right then down from the
		top-left corner of the text display area, and starting from 0@0
!

!CharacterBlock methodsFor: 'comparing' stamp: ''!
< aCharacterBlock 
	"Answer whether the string index of the receiver precedes that of 
	aCharacterBlock."

	^stringIndex < aCharacterBlock stringIndex! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
<= aCharacterBlock 
	"Answer whether the string index of the receiver does not come after that 
	of aCharacterBlock."

	^(self > aCharacterBlock) not! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
= aCharacterBlock

	self species = aCharacterBlock species
		ifTrue: [^stringIndex = aCharacterBlock stringIndex]
		ifFalse: [^false]! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
> aCharacterBlock 
	"Answer whether the string index of the receiver comes after that of 
	aCharacterBlock."

	^aCharacterBlock < self! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
>= aCharacterBlock 
	"Answer whether the string index of the receiver does not precede that of 
	aCharacterBlock."

	^(self < aCharacterBlock) not! !

!CharacterBlock methodsFor: 'comparing' stamp: 'th 9/17/2002 11:54'!
max: aCharacterBlock
	aCharacterBlock ifNil:[^self].
	^aCharacterBlock > self
		ifTrue:[ aCharacterBlock]
		ifFalse:[self].! !

!CharacterBlock methodsFor: 'comparing' stamp: 'th 9/17/2002 11:54'!
min: aCharacterBlock
	aCharacterBlock ifNil:[^self].
	^aCharacterBlock < self
		ifTrue:[ aCharacterBlock]
		ifFalse:[self].! !

!CharacterBlock methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
moveBy: aPoint 
	"Change the corner positions of the receiver so that its area translates by 
	the amount defined by the argument, aPoint."
	origin := origin + aPoint.
	corner := corner + aPoint! !

!CharacterBlock methodsFor: 'printing' stamp: 'di 12/2/97 19:15'!
printOn: aStream

	aStream nextPutAll: 'a CharacterBlock with index '.
	stringIndex printOn: aStream.
	(text ~~ nil and: [text size> 0 and: [stringIndex between: 1 and: text size]])
		ifTrue: [aStream nextPutAll: ' and character '.
				(text at: stringIndex) printOn: aStream].
	aStream nextPutAll: ' and rectangle '.
	super printOn: aStream.
	textLine ifNotNil: [aStream cr; nextPutAll: ' in '.
				textLine printOn: aStream].
! !

!CharacterBlock methodsFor: 'accessing' stamp: ''!
stringIndex
	"Answer the position of the receiver in the string it indexes."

	^stringIndex! !

!CharacterBlock methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
stringIndex: anInteger text: aText topLeft: topLeft extent: extent 
	stringIndex := anInteger.
	text := aText.
	super 
		setOrigin: topLeft
		corner: topLeft + extent! !

!CharacterBlock methodsFor: 'accessing' stamp: 'di 12/2/97 14:33'!
textLine
	^ textLine! !

!CharacterBlock methodsFor: 'accessing' stamp: 'lr 7/4/2009 10:42'!
textLine: aLine 
	textLine := aLine! !

Object subclass: #CharacterScanner
	instanceVariableNames: 'destX lastIndex xTable destY stopConditions text textStyle alignment leftMargin rightMargin font line runStopIndex spaceCount spaceWidth emphasisCode kern indentationLevel wantsColumnBreaks pendingKernX'
	classVariableNames: 'DefaultStopConditions NilCondition PaddedSpaceCondition SpaceCondition'
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CharacterScanner commentStamp: '<historical>' prior: 0!
My instances hold the state associated with scanning text. My subclasses scan characters for specified purposes, such as computing a CharacterBlock or placing characters into Forms.!

CharacterScanner subclass: #CharacterBlockScanner
	instanceVariableNames: 'characterPoint characterIndex lastCharacter lastCharacterExtent lastSpaceOrTabExtent nextLeftMargin specialWidth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Text'!

!CharacterBlockScanner commentStamp: '<historical>' prior: 0!
My instances are used to scan text to compute the CharacterBlock for a character specified by its index in the text or its proximity to the cursor location.!

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
buildCharacterBlockIn: para 
	"handle nullText"
	| lineIndex runLength lineStop done stopCondition |
	(para numberOfLines = 0 or: [ text size = 0 ]) ifTrue: 
		[ ^ CharacterBlock new 
			stringIndex: 1
			text: para text
			topLeft: (para 
				leftMarginForDisplayForLine: 1
				alignment: (alignment ifNil: 
					[ "like being off end of string"
					textStyle alignment ])) @ para compositionRectangle top
			extent: 0 @ textStyle lineGrid ].
	"find the line"
	lineIndex := para lineIndexOfTop: characterPoint y.
	destY := para topAtLineIndex: lineIndex.
	line := para lines at: lineIndex.
	lastIndex := line first.
	rightMargin := para rightMarginForDisplay.
	self setStopConditions.	" also loads the font and loads all emphasis attributes "
	(lineIndex = para numberOfLines and: [ destY + line lineHeight < characterPoint y ]) 
		ifTrue: 
			[ "if beyond lastLine, force search to last character"
			self characterPointSetX: rightMargin ]
		ifFalse: 
			[ characterPoint y < para compositionRectangle top ifTrue: 
				[ "force search to first line"
				characterPoint := para compositionRectangle topLeft ].
			characterPoint x > rightMargin ifTrue: [ self characterPointSetX: rightMargin ] ].
	destX := leftMargin := para 
		leftMarginForDisplayForLine: lineIndex
		alignment: (alignment ifNil: [ textStyle alignment ]).
	nextLeftMargin := para 
		leftMarginForDisplayForLine: lineIndex + 1
		alignment: (alignment ifNil: [ textStyle alignment ]).
	lastIndex := line first.
	self setStopConditions.	"also sets font"
	runLength := text runLengthFor: line first.
	characterIndex == nil 
		ifTrue: [ lineStop := line last	"characterBlockAtPoint" ]
		ifFalse: [ lineStop := characterIndex	"characterBlockForIndex" ].
	(runStopIndex := lastIndex + (runLength - 1)) > lineStop ifTrue: [ runStopIndex := lineStop ].
	lastCharacterExtent := 0 @ line lineHeight.
	spaceCount := 0.
	done := false.
	self handleIndentation.
	[ done ] whileFalse: 
		[ stopCondition := self 
			scanCharactersFrom: lastIndex
			to: runStopIndex
			in: text string
			rightX: characterPoint x
			stopConditions: stopConditions
			kern: kern.

		"see setStopConditions for stopping conditions for character block 	operations."
		self lastCharacterExtentSetX: (font widthOf: (text at: lastIndex)).
		(self perform: stopCondition) ifTrue: 
			[ characterIndex == nil 
				ifTrue: 
					[ "characterBlockAtPoint"
					^ CharacterBlock new 
						stringIndex: lastIndex
						text: text
						topLeft: characterPoint + (font descentKern @ 0)
						extent: lastCharacterExtent ]
				ifFalse: 
					[ "characterBlockForIndex"
					^ CharacterBlock new 
						stringIndex: lastIndex
						text: text
						topLeft: characterPoint + ((font descentKern - kern) @ 0)
						extent: lastCharacterExtent ] ] ]! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'nk 11/22/2004 14:32'!
characterBlockAtPoint: aPoint index: index in: textLine
	"This method is the Morphic characterBlock finder.  It combines
	MVC's characterBlockAtPoint:, -ForIndex:, and buildCharcterBlock:in:"
	| runLength lineStop done stopCondition |
	line := textLine.
	rightMargin := line rightMargin.
	lastIndex := line first.
	self setStopConditions.		"also sets font"
	characterIndex := index.  " == nil means scanning for point"
	characterPoint := aPoint.
	(characterPoint isNil or: [characterPoint y > line bottom])
		ifTrue: [characterPoint := line bottomRight].
	(text isEmpty or: [(characterPoint y < line top or: [characterPoint x < line left])
				or: [characterIndex notNil and: [characterIndex < line first]]])
		ifTrue:	[^ (CharacterBlock new stringIndex: line first text: text
					topLeft: line leftMargin@line top extent: 0 @ textStyle lineGrid)
					textLine: line].
	destX := leftMargin := line leftMarginForAlignment: alignment.
	destY := line top.
	runLength := text runLengthFor: line first.
	characterIndex
		ifNotNil:	[lineStop := characterIndex  "scanning for index"]
		ifNil:	[lineStop := line last  "scanning for point"].
	runStopIndex := lastIndex + (runLength - 1) min: lineStop.
	lastCharacterExtent := 0 @ line lineHeight.
	spaceCount := 0.

	done  := false.
	[done] whileFalse:
		[stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
			in: text string rightX: characterPoint x
			stopConditions: stopConditions kern: kern.
		"see setStopConditions for stopping conditions for character block 	operations."
		self lastCharacterExtentSetX: (specialWidth
			ifNil: [font widthOf: (text at: lastIndex)]
			ifNotNil: [specialWidth]).
		(self perform: stopCondition) ifTrue:
			[characterIndex
				ifNil: [
					"Result for characterBlockAtPoint: "
					(stopCondition ~~ #cr and: [ lastIndex == line last
						and: [ aPoint x > ((characterPoint x) + (lastCharacterExtent x / 2)) ]])
							ifTrue: [ "Correct for right half of last character in line"
								^ (CharacterBlock new stringIndex: lastIndex + 1
										text: text
										topLeft: characterPoint + (lastCharacterExtent x @ 0) + (font descentKern @ 0)
										extent:  0 @ lastCharacterExtent y)
									textLine: line ].
						^ (CharacterBlock new stringIndex: lastIndex
							text: text topLeft: characterPoint + (font descentKern @ 0)
							extent: lastCharacterExtent - (font baseKern @ 0))
									textLine: line]
				ifNotNil: ["Result for characterBlockForIndex: "
						^ (CharacterBlock new stringIndex: characterIndex
							text: text topLeft: characterPoint + ((font descentKern) - kern @ 0)
							extent: lastCharacterExtent)
									textLine: line]]]! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
characterPointSetX: xVal 
	characterPoint := xVal @ characterPoint y! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'nice 11/30/2009 23:04'!
cr
	"Answer a CharacterBlock that specifies the current location of the mouse 
	relative to a carriage return stop condition that has just been 
	encountered. The ParagraphEditor convention is to denote selections by 
	CharacterBlocks, sometimes including the carriage return (cursor is at 
	the end) and sometimes not (cursor is in the middle of the text)."
	((characterIndex ~= nil and: [ characterIndex > text size ]) or: 
		[ line last = text size and: [ destY + line lineHeight < characterPoint y ] ]) ifTrue: 
		[ "When off end of string, give data for next character"
		destY := destY + line lineHeight.
		lastCharacter := nil.
		characterPoint := (nextLeftMargin ifNil: [ leftMargin ]) @ destY.
		(lastIndex < text size and: [(text at: lastIndex) = CR and: [(text at: lastIndex+1) = Character lf]])
			ifTrue: [lastIndex := lastIndex + 2]
			ifFalse: [lastIndex := lastIndex + 1].
		self lastCharacterExtentSetX: 0.
		^ true ].
	lastCharacter := CR.
	characterPoint := destX @ destY.
	self lastCharacterExtentSetX: rightMargin - destX.
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 12:49'!
crossedX
	"Text display has wrapping. The scanner just found a character past the x 
	location of the cursor. We know that the cursor is pointing at a character 
	or before one."

	| leadingTab currentX |
	characterIndex == nil ifFalse: [
		"If the last character of the last line is a space,
		and it crosses the right margin, then locating
		the character block after it is impossible without this hack."
		characterIndex > text size ifTrue: [
			lastIndex := characterIndex.
			characterPoint := (nextLeftMargin ifNil: [leftMargin]) @ (destY + line lineHeight).
			^true]].
	characterPoint x <= (destX + (lastCharacterExtent x // 2))
		ifTrue:	[lastCharacter := (text at: lastIndex).
				characterPoint := destX @ destY.
				^true].
	lastIndex >= line last 
		ifTrue:	[lastCharacter := (text at: line last).
				characterPoint := destX @ destY.
				^true].

	"Pointing past middle of a character, return the next character."
	lastIndex := lastIndex + 1.
	lastCharacter := text at: lastIndex.
	currentX := destX + lastCharacterExtent x + kern.
	self lastCharacterExtentSetX: (font widthOf: lastCharacter).
	characterPoint := currentX @ destY.
	lastCharacter = Space ifFalse: [^ true].

	"Yukky if next character is space or tab."
	alignment = Justified ifTrue:
		[self lastCharacterExtentSetX:
			(lastCharacterExtent x + 	(line justifiedPadFor: (spaceCount + 1) font: font)).
		^ true].

	true ifTrue: [^ true].
	"NOTE:  I find no value to the following code, and so have defeated it - DI"

	"See tabForDisplay for illumination on the following awfulness."
	leadingTab := true.
	line first to: lastIndex - 1 do:
		[:index | (text at: index) ~= Tab ifTrue: [leadingTab := false]].
	(alignment ~= Justified or: [leadingTab])
		ifTrue:	[self lastCharacterExtentSetX: (textStyle nextTabXFrom: currentX
					leftMargin: leftMargin rightMargin: rightMargin) -
						currentX]
		ifFalse:	[self lastCharacterExtentSetX:  (((currentX + (textStyle tabWidth -
						(line justifiedTabDeltaFor: spaceCount))) -
							currentX) max: 0)].
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
endOfRun
	"Before arriving at the cursor location, the selection has encountered an 
	end of run. Answer false if the selection continues, true otherwise. Set 
	up indexes for building the appropriate CharacterBlock."
	| runLength lineStop |
	(((characterIndex ~~ nil and: [ runStopIndex < characterIndex and: [ runStopIndex < text size ] ]) or: [ characterIndex == nil and: [ lastIndex < line last ] ]) or: 
		[ lastIndex < line last and: 
			[ (text at: lastIndex) leadingChar ~= (text at: lastIndex + 1) leadingChar and: [ lastIndex ~= characterIndex ] ] ]) ifTrue: 
		[ "We're really at the end of a real run."
		runLength := text runLengthFor: (lastIndex := lastIndex + 1).
		characterIndex ~~ nil 
			ifTrue: [ lineStop := characterIndex	"scanning for index" ]
			ifFalse: [ lineStop := line last	"scanning for point" ].
		(runStopIndex := lastIndex + (runLength - 1)) > lineStop ifTrue: [ runStopIndex := lineStop ].
		self setStopConditions.
		^ false ].
	lastCharacter := text at: lastIndex.
	characterPoint := destX @ destY.
	((lastCharacter = Space and: [ alignment = Justified ]) or: [ lastCharacter = Tab and: [ lastSpaceOrTabExtent notNil ] ]) ifTrue: [ lastCharacterExtent := lastSpaceOrTabExtent ].
	characterIndex ~~ nil ifTrue: 
		[ "If scanning for an index and we've stopped on that index,
				then we back destX off by the width of the character stopped on
				(it will be pointing at the right side of the character) and return"
		runStopIndex = characterIndex ifTrue: 
			[ self characterPointSetX: destX - lastCharacterExtent x.
			^ true ].
		"Otherwise the requested index was greater than the length of the
				string.  Return string size + 1 as index, indicate further that off the
				string by setting character to nil and the extent to 0."
		lastIndex := lastIndex + 1.
		lastCharacter := nil.
		self lastCharacterExtentSetX: 0.
		^ true ].

	"Scanning for a point and either off the end of the line or off the end of the string."
	runStopIndex = text size ifTrue: 
		[ "off end of string"
		lastIndex := lastIndex + 1.
		lastCharacter := nil.
		self lastCharacterExtentSetX: 0.
		^ true ].
	"just off end of line without crossing x"
	lastIndex := lastIndex + 1.
	^ true! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
indentationLevel: anInteger 
	super indentationLevel: anInteger.
	nextLeftMargin := leftMargin.
	indentationLevel timesRepeat: 
		[ nextLeftMargin := textStyle 
			nextTabXFrom: nextLeftMargin
			leftMargin: leftMargin
			rightMargin: rightMargin ]! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
lastCharacterExtentSetX: xVal 
	lastCharacterExtent := xVal @ lastCharacterExtent y! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
lastSpaceOrTabExtentSetX: xVal 
	lastSpaceOrTabExtent := xVal @ lastSpaceOrTabExtent y! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 12:50'!
paddedSpace
	"When the line is justified, the spaces will not be the same as the font's 
	space character. A padding of extra space must be considered in trying 
	to find which character the cursor is pointing at. Answer whether the 
	scanning has crossed the cursor."

	| pad |
	pad := 0.
	spaceCount := spaceCount + 1.
	pad := line justifiedPadFor: spaceCount font: font.
	lastSpaceOrTabExtent := lastCharacterExtent copy.
	self lastSpaceOrTabExtentSetX:  spaceWidth + pad.
	(destX + lastSpaceOrTabExtent x)  >= characterPoint x
		ifTrue: [lastCharacterExtent := lastSpaceOrTabExtent copy.
				^self crossedX].
	lastIndex := lastIndex + 1.
	destX := destX + lastSpaceOrTabExtent x.
	^ false
! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
placeEmbeddedObject: anchoredMorph 
	"Workaround: The following should really use #textAnchorType"
	anchoredMorph relativeTextAnchorPosition ifNotNil: [ ^ true ].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: [ ^ false ].
	specialWidth := anchoredMorph width.
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
setFont
	specialWidth := nil.
	super setFont! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'ar 10/18/2004 14:30'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont.
	self setConditionArray: (alignment = Justified ifTrue: [#paddedSpace]).
! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'nice 4/19/2011 00:20'!
tab
	| currentX |
	currentX := (alignment = Justified and: [ self leadingTab not ]) 
		ifTrue: 
			[ "imbedded tabs in justified text are weird"
			destX + (textStyle tabWidth - (line justifiedTabDeltaFor: spaceCount)) max: destX ]
		ifFalse: 
			[ textStyle 
				nextTabXFrom: destX
				leftMargin: leftMargin
				rightMargin: rightMargin ].
	lastSpaceOrTabExtent := lastCharacterExtent copy.
	self lastSpaceOrTabExtentSetX: (currentX - destX max: 0).
	currentX >= characterPoint x ifTrue: 
		[ lastCharacterExtent := lastSpaceOrTabExtent copy.
		^ self crossedX ].
	destX := currentX.
	lastIndex := lastIndex + 1.
	^ false! !

!CharacterScanner class methodsFor: 'class initialization' stamp: 'nice 3/8/2010 11:51'!
initialize
"
	CharacterScanner initialize
"
	| a |
	a := TextStopConditions new.
	a at: 1 + 1 put: #embeddedObject.
	a at: Tab asciiValue + 1 put: #tab.
	a at: CR asciiValue + 1 put: #cr.
	a at: Character lf asciiValue + 1 put: #cr.
	
	NilCondition := a copy.
	DefaultStopConditions := a copy.

	PaddedSpaceCondition := a copy.
	PaddedSpaceCondition at: Space asciiValue + 1 put: #paddedSpace.
	
	SpaceCondition := a copy.
	SpaceCondition at: Space asciiValue + 1 put: #space.
! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
addEmphasis: code 
	"Set the bold-ital-under-strike emphasis."
	emphasisCode := emphasisCode bitOr: code! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
addKern: kernDelta 
	"Set the current kern amount."
	kern := kern + kernDelta! !

!CharacterScanner methodsFor: 'scanning' stamp: 'nice 2/28/2010 18:17'!
basicScanCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta
	"Primitive. This is the inner loop of text display--but see 
	scanCharactersFrom: to:rightX: which would get the string, 
	stopConditions and displaying from the instance. March through source 
	String from startIndex to stopIndex. If any character is flagged with a 
	non-nil entry in stops, then return the corresponding value. Determine 
	width of each character from xTable, indexed by map. 
	If dextX would exceed rightX, then return stops at: 258. 
	Advance destX by the width of the character. If stopIndex has been
	reached, then return stops at: 257. Optional. 
	See Object documentation whatIsAPrimitive."
	| ascii nextDestX char  floatDestX widthAndKernedWidth nextChar atEndOfRun |
	<primitive: 103>
	lastIndex := startIndex.
	floatDestX := destX.
	widthAndKernedWidth := Array new: 2.
	atEndOfRun := false.
	[lastIndex <= stopIndex]
		whileTrue: 
			[char := (sourceString at: lastIndex).
			ascii := char asciiValue + 1.
			(stops at: ascii) == nil ifFalse: [^stops at: ascii].
			"Note: The following is querying the font about the width
			since the primitive may have failed due to a non-trivial
			mapping of characters to glyphs or a non-existing xTable."
			nextChar := (lastIndex + 1 <= stopIndex) 
				ifTrue:[sourceString at: lastIndex + 1]
				ifFalse:[
					atEndOfRun := true.
					"if there is a next char in sourceString, then get the kern 
					and store it in pendingKernX"
					lastIndex + 1 <= sourceString size
						ifTrue:[sourceString at: lastIndex + 1]
						ifFalse:[	nil]].
			font 
				widthAndKernedWidthOfLeft: char 
				right: nextChar
				into: widthAndKernedWidth.
			nextDestX := floatDestX + (widthAndKernedWidth at: 1).
			nextDestX > rightX ifTrue: [^stops crossedX].
			floatDestX := floatDestX + kernDelta + (widthAndKernedWidth at: 2).
			atEndOfRun 
				ifTrue:[
					pendingKernX := (widthAndKernedWidth at: 2) - (widthAndKernedWidth at: 1).
					floatDestX := floatDestX - pendingKernX].
			destX := floatDestX.
			lastIndex := lastIndex + 1].
	lastIndex := stopIndex.
	^stops endOfRun! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'tween 4/6/2007 11:16'!
columnBreak

	pendingKernX := 0.
	^true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
embeddedObject
	| savedIndex |
	savedIndex := lastIndex.
	text 
		attributesAt: lastIndex
		do: 
			[ :attr | 
			attr anchoredMorph ifNotNil: 
				[ "Following may look strange but logic gets reversed.
			If the morph fits on this line we're not done (return false for true) 
			and if the morph won't fit we're done (return true for false)"
				(self placeEmbeddedObject: attr anchoredMorph) ifFalse: [ ^ true ] ] ].
	lastIndex := savedIndex + 1.	"for multiple(!!) embedded morphs"
	^ false! !

!CharacterScanner methodsFor: 'scanning' stamp: 'hmm 7/15/2000 22:40'!
handleIndentation
	self indentationLevel timesRepeat: [
		self plainTab]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 5/17/2000 18:20'!
indentationLevel
	"return the number of tabs that are currently being placed at the beginning of each line"
	^indentationLevel ifNil:[0]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
indentationLevel: anInteger 
	"set the number of tabs to put at the beginning of each line"
	indentationLevel := anInteger! !

!CharacterScanner methodsFor: 'initialize' stamp: 'alain.plantec 5/28/2009 09:44'!
initialize
	super initialize.
	destX := destY := leftMargin := 0.! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
initializeFromParagraph: aParagraph clippedBy: clippingRectangle 
	text := aParagraph text.
	textStyle := aParagraph textStyle! !

!CharacterScanner methodsFor: 'initialize' stamp: 'ul 3/8/2010 04:44'!
initializeStringMeasurer

	stopConditions := TextStopConditions new
! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'yo 12/18/2002 12:32'!
isBreakableAtIndex: index

	^ (EncodedCharSet at: ((text at: index) leadingChar + 1)) isBreakableAt: index in: text.
! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 1/8/2000 14:23'!
leadingTab
	"return true if only tabs lie to the left"
	line first to: lastIndex do:
		[:i | (text at: i) == Tab ifFalse: [^ false]].
	^ true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
measureString: aString inFont: aFont from: startIndex to: stopIndex 
	"WARNING: In order to use this method the receiver has to be set up using #initializeStringMeasurer"
	destX := destY := lastIndex := 0.
	xTable := aFont xTable.
	self 
		scanCharactersFrom: startIndex
		to: stopIndex
		in: aString
		rightX: 999999
		stopConditions: stopConditions
		kern: 0.
	^ destX! !

!CharacterScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
placeEmbeddedObject: anchoredMorph 
	"Place the anchoredMorph or return false if it cannot be placed.
	In any event, advance destX by its width."
	"Workaround: The following should really use #textAnchorType"
	| w |
	anchoredMorph relativeTextAnchorPosition ifNotNil: [ ^ true ].
	destX := destX + (w := anchoredMorph width).
	(destX > rightMargin and: [ leftMargin + w <= rightMargin ]) ifTrue: 
		[ "Won't fit, but would on next line"
		^ false ].
	lastIndex := lastIndex + 1.
	self setFont.	"Force recalculation of emphasis for next run"
	^ true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'nice 3/16/2010 20:38'!
plainTab
	"This is the basic method of adjusting destX for a tab."
	destX := (alignment = Justified and: [ self leadingTab not ]) 
		ifTrue: 
			[ "embedded tabs in justified text are weird"
			destX + (textStyle tabWidth - (line justifiedTabDeltaFor: spaceCount)) max: destX ]
		ifFalse: 
			[ textStyle 
				nextTabXFrom: destX
				leftMargin: leftMargin
				rightMargin: rightMargin ]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'nice 3/6/2010 14:47'!
scanCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta 
	| startEncoding selector |
	sourceString isByteString ifTrue: 
		[ ^ self 
			basicScanCharactersFrom: startIndex
			to: stopIndex
			in: sourceString
			rightX: rightX
			stopConditions: stops
			kern: kernDelta ].
	sourceString isWideString ifTrue: 
		[ startIndex > stopIndex ifTrue: 
			[ lastIndex := stopIndex.
			^ stops endOfRun ].
		startEncoding := (sourceString at: startIndex) leadingChar.
		selector := (EncodedCharSet charsetAt: startEncoding) scanSelector.
		^ self 
			perform: selector
			withArguments: (Array 
					with: startIndex
					with: stopIndex
					with: sourceString
					with: rightX
					with: stops
					with: kernDelta) ].
	^ stops endOfRun! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
scanJapaneseCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta 
	| ascii encoding f nextDestX maxAscii startEncoding |
	lastIndex := startIndex.
	lastIndex > stopIndex ifTrue: 
		[ lastIndex := stopIndex.
		^ stops endOfRun ].
	startEncoding := (sourceString at: startIndex) leadingChar.
	font ifNil: [ font := (TextSharedInformation at: #DefaultMultiStyle) fontArray at: 1 ].
	((font isMemberOf: StrikeFontSet) or: [ font isKindOf: TTCFontSet ]) 
		ifTrue: 
			[ maxAscii := font maxAsciiFor: startEncoding.
			f := font fontArray at: startEncoding + 1.
			"xTable := f xTable.
		maxAscii := xTable size - 2."
			spaceWidth := f widthOf: Space ]
		ifFalse: [ maxAscii := font maxAscii ].
	[ lastIndex <= stopIndex ] whileTrue: 
		[ encoding := (sourceString at: lastIndex) leadingChar.
		encoding ~= startEncoding ifTrue: 
			[ lastIndex := lastIndex - 1.
			^ stops endOfRun ].
		ascii := (sourceString at: lastIndex) charCode.
		ascii > maxAscii ifTrue: [ ascii := maxAscii ].
		(encoding = 0 and: [ascii < stops size and: [ (stopConditions at: ascii + 1) ~~ nil ]]) ifTrue: [ ^ stops at: ascii + 1 ].
		nextDestX := destX + (font widthOf: (sourceString at: lastIndex)).
		nextDestX > rightX ifTrue: [ ^ stops crossedX ].
		destX := nextDestX + kernDelta.
		"destX printString displayAt: 0@(lastIndex*20)."
		lastIndex := lastIndex + 1 ].
	lastIndex := stopIndex.
	^ stops endOfRun! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'nice 2/28/2010 19:10'!
scanMultiCharactersCombiningFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta
	"Implement a workaround for unicode composing.
	a MultiCharacterScanner should better be used to handle combination."

	^self scanMultiCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'BenjaminVanRyseghem 11/28/2010 19:02'!
scanMultiCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta 
	| ascii encoding f nextDestX maxAscii startEncoding |
	lastIndex := startIndex.
	lastIndex > stopIndex ifTrue: 
		[ lastIndex := stopIndex.
		^ stops endOfRun ].
	startEncoding := (sourceString at: startIndex) leadingChar.
	font ifNil: [ font := (TextSharedInformation at: #DefaultMultiStyle) fontArray at: 1 ].
	((font isMemberOf: StrikeFontSet) or: [ font isKindOf: TTCFontSet ]) 
		ifTrue: 
			[ maxAscii := font maxAsciiFor: startEncoding.
			f := font fontArray at: startEncoding + 1.
			spaceWidth := f widthOf: Space ]
		ifFalse: [ maxAscii := font maxAscii ].
	[ lastIndex <= stopIndex ] whileTrue: 
		[ encoding := (sourceString at: lastIndex) leadingChar.
		encoding ~= startEncoding ifTrue: 
			[ lastIndex := lastIndex - 1.
			^ stops endOfRun ].
		ascii := (sourceString at: lastIndex) charCode.
		ascii > maxAscii ifTrue: [ ascii := maxAscii ].
		(encoding = 0 and: [ascii < stops size and: [ (stopConditions at: ascii + 1) ~~ nil ]]) ifTrue: [ ^ stops at: ascii + 1 ].
		nextDestX := destX + (font widthOf: (sourceString at: lastIndex)).
		nextDestX > rightX ifTrue: [ ^ stops crossedX ].
		destX := nextDestX + kernDelta.
		"destX printString displayAt: 0@(lastIndex*20)."
		lastIndex := lastIndex + 1 ].
	lastIndex := stopIndex.
	^ stops endOfRun! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setActualFont: aFont 
	"Set the basal font to an isolated font reference."
	font := aFont! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setAlignment: style 
	alignment := style! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
setConditionArray: aSymbol 
	aSymbol == #paddedSpace ifTrue: [ ^ stopConditions := PaddedSpaceCondition copy ].
	aSymbol == #space ifTrue: [ ^ stopConditions := SpaceCondition copy ].
	aSymbol == nil ifTrue: [ ^ stopConditions := NilCondition copy ].
	self error: 'undefined stopcondition for space character'! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'tween 4/6/2007 09:59'!
setFont
	| priorFont |
	"Set the font and other emphasis."
	priorFont := font.
	text == nil ifFalse:[
		emphasisCode := 0.
		kern := 0.
		indentationLevel := 0.
		alignment := textStyle alignment.
		font := nil.
		(text attributesAt: lastIndex forStyle: textStyle)
			do: [:att | att emphasizeScanner: self]].
	font == nil ifTrue:
		[self setFont: textStyle defaultFontIndex].
	font := font emphasized: emphasisCode.
	priorFont 
		ifNotNil: [
			font = priorFont 
				ifTrue:[
					"font is the same, perhaps the color has changed?
					We still want kerning between chars of the same
					font, but of different color. So add any pending kern to destX"
					destX := destX + (pendingKernX ifNil:[0])].
			destX := destX + priorFont descentKern].
	pendingKernX := 0. "clear any pending kern so there is no danger of it being added twice"
	destX := destX - font descentKern.
	"NOTE: next statement should be removed when clipping works"
	leftMargin ifNotNil: [destX := destX max: leftMargin].
	kern := kern - font baseKern.

	"Install various parameters from the font."
	spaceWidth := font widthOf: Space.
	xTable := font xTable.
	stopConditions := DefaultStopConditions.! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:28'!
setFont: fontNumber
	"Set the font by number from the textStyle."

	self setActualFont: (textStyle fontAt: fontNumber)! !

!CharacterScanner methodsFor: 'private' stamp: 'lr 7/4/2009 10:42'!
text: t textStyle: ts 
	text := t.
	textStyle := ts! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:28'!
textColor: ignored
	"Overridden in DisplayScanner"! !

!CharacterScanner methodsFor: 'initialize' stamp: 'lr 7/4/2009 10:42'!
wantsColumnBreaks: aBoolean 
	wantsColumnBreaks := aBoolean! !

CharacterScanner subclass: #CompositionScanner
	instanceVariableNames: 'spaceX spaceIndex lineHeight baseline lineHeightAtSpace baselineAtSpace'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CompositionScanner commentStamp: '<historical>' prior: 0!
CompositionScanners are used to measure text and determine where line breaks and space padding should occur.!

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:13'!
columnBreak

	"Answer true. Set up values for the text line interval currently being 
	composed."

	pendingKernX := 0.
	line stop: lastIndex.
	spaceX := destX.
	line paddingWidth: rightMargin - spaceX.
	^true! !

!CompositionScanner methodsFor: 'scanning' stamp: 'nice 3/16/2010 20:42'!
composeFrom: startIndex inRectangle: lineRectangle firstLine: firstLine leftSide: leftSide rightSide: rightSide 
	"Answer an instance of TextLineInterval that represents the next line in the paragraph."
	"Set up margins"
	| runLength stopCondition |
	leftMargin := lineRectangle left.
	leftSide ifTrue: 
		[ leftMargin := leftMargin + (firstLine 
				ifTrue: [ textStyle firstIndent ]
				ifFalse: [ textStyle restIndent ]) ].
	destX := spaceX := leftMargin.
	rightMargin := lineRectangle right.
	rightSide ifTrue: [ rightMargin := rightMargin - textStyle rightIndent ].
	lastIndex := startIndex.	"scanning sets last index"
	destY := lineRectangle top.
	lineHeight := baseline := 0.	"Will be increased by setFont"
	self setStopConditions.	"also sets font"
	runLength := text runLengthFor: startIndex.
	runStopIndex := (lastIndex := startIndex) + (runLength - 1).
	line := (TextLine 
		start: lastIndex
		stop: 0
		internalSpaces: 0
		paddingWidth: 0) rectangle: lineRectangle.
	spaceCount := 0.
	self handleIndentation.
	leftMargin := destX.
	line leftMargin: leftMargin.
	[
		stopCondition := self 
			scanCharactersFrom: lastIndex
			to: runStopIndex
			in: text string
			rightX: rightMargin
			stopConditions: stopConditions
			kern: kern.
		"See setStopConditions for stopping conditions for composing."
		self perform: stopCondition
	] whileFalse.
	^ line
		lineHeight: lineHeight + textStyle leading
		baseline: baseline + textStyle leading! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'nice 11/16/2009 09:23'!
cr
	"Answer true. Set up values for the text line interval currently being 
	composed."

	pendingKernX := 0.
	(lastIndex < text size and: [(text at: lastIndex) = CR and: [(text at: lastIndex+1) = Character lf]]) ifTrue: [lastIndex := lastIndex + 1].
	line stop: lastIndex.
	spaceX := destX.
	line paddingWidth: rightMargin - spaceX.
	^true! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:14'!
crossedX
	"There is a word that has fallen across the right edge of the composition 
	rectangle. This signals the need for wrapping which is done to the last 
	space that was encountered, as recorded by the space stop condition."

	pendingKernX := 0.
	spaceCount >= 1 ifTrue:
		["The common case. First back off to the space at which we wrap."
		line stop: spaceIndex.
		lineHeight := lineHeightAtSpace.
		baseline := baselineAtSpace.
		spaceCount := spaceCount - 1.
		spaceIndex := spaceIndex - 1.

		"Check to see if any spaces preceding the one at which we wrap.
			Double space after punctuation, most likely."
		[(spaceCount > 1 and: [(text at: spaceIndex) = Space])]
			whileTrue:
				[spaceCount := spaceCount - 1.
				"Account for backing over a run which might
					change width of space."
				font := text fontAt: spaceIndex withStyle: textStyle.
				spaceIndex := spaceIndex - 1.
				spaceX := spaceX - (font widthOf: Space)].
		line paddingWidth: rightMargin - spaceX.
		line internalSpaces: spaceCount]
	ifFalse:
		["Neither internal nor trailing spaces -- almost never happens."
		lastIndex := lastIndex - 1.
		[destX <= rightMargin]
			whileFalse:
				[destX := destX - (font widthOf: (text at: lastIndex)).
				lastIndex := lastIndex - 1].
		spaceX := destX.
		line paddingWidth: rightMargin - destX.
		line stop: (lastIndex max: line first)].
	^true! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
endOfRun
	"Answer true if scanning has reached the end of the paragraph. 
	Otherwise step conditions (mostly install potential new font) and answer 
	false."
	| runLength |
	lastIndex = text size 
		ifTrue: 
			[ line stop: lastIndex.
			spaceX := destX.
			line paddingWidth: rightMargin - destX.
			^ true ]
		ifFalse: 
			[ runLength := text runLengthFor: (lastIndex := lastIndex + 1).
			runStopIndex := lastIndex + (runLength - 1).
			self setStopConditions.
			^ false ]! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
placeEmbeddedObject: anchoredMorph 
	"Workaround: The following should really use #textAnchorType"
	| descent |
	anchoredMorph relativeTextAnchorPosition ifNotNil: [ ^ true ].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: 
		[ "It doesn't fit"
		"But if it's the first character then leave it here"
		lastIndex < line first ifFalse: 
			[ line stop: lastIndex - 1.
			^ false ] ].
	descent := lineHeight - baseline.
	lineHeight := lineHeight max: anchoredMorph height.
	baseline := lineHeight - descent.
	line stop: lastIndex.
	^ true! !

!CompositionScanner methodsFor: 'accessing' stamp: 'ar 1/8/2000 14:35'!
rightX
	"Meaningful only when a line has just been composed -- refers to the 
	line most recently composed. This is a subtrefuge to allow for easy 
	resizing of a composition rectangle to the width of the maximum line. 
	Useful only when there is only one line in the form or when each line 
	is terminated by a carriage return. Handy for sizing menus and lists."

	^spaceX! !

!CompositionScanner methodsFor: 'scanning' stamp: 'lr 7/4/2009 10:42'!
setActualFont: aFont 
	"Keep track of max height and ascent for auto lineheight"
	| descent |
	super setActualFont: aFont.
	lineHeight == nil 
		ifTrue: 
			[ descent := font descent.
			baseline := font ascent.
			lineHeight := baseline + descent ]
		ifFalse: 
			[ descent := lineHeight - baseline max: font descent.
			baseline := baseline max: font ascent.
			lineHeight := lineHeight max: baseline + descent ]! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'lr 7/4/2009 10:42'!
setFont
	super setFont.
	stopConditions == DefaultStopConditions ifTrue: [ stopConditions := stopConditions copy ].
	stopConditions 
		at: Space asciiValue + 1
		put: #space.
	wantsColumnBreaks == true ifTrue: 
		[ stopConditions 
			at: TextComposer characterForColumnBreak asciiValue + 1
			put: #columnBreak ]! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/8/2000 14:37'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:14'!
space
	"Record left x and character index of the space character just encounted. 
	Used for wrap-around. Answer whether the character has crossed the 
	right edge of the composition rectangle of the paragraph."

	pendingKernX := 0.
	spaceX := destX.
	destX := spaceX + spaceWidth.
	spaceIndex := lastIndex.
	lineHeightAtSpace := lineHeight.
	baselineAtSpace := baseline.
	lastIndex := lastIndex + 1.
	spaceCount := spaceCount + 1.
	destX > rightMargin ifTrue: 	[^self crossedX].
	^false
! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'tween 4/6/2007 11:15'!
tab
	"Advance destination x according to tab settings in the paragraph's 
	textStyle. Answer whether the character has crossed the right edge of 
	the composition rectangle of the paragraph."

	pendingKernX := 0.
	destX := textStyle
				nextTabXFrom: destX leftMargin: leftMargin rightMargin: rightMargin.
	destX > rightMargin ifTrue:	[^self crossedX].
	lastIndex := lastIndex + 1.
	^false
! !

Object subclass: #StandardFonts
	instanceVariableNames: ''
	classVariableNames: 'ButtonFont CodeFont HaloFont ListFont MenuFont WindowTitleFont'
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!StandardFonts commentStamp: 'LaurentLaffont 2/10/2011 18:19' prior: 0!
I am a facade and a global reference to actual fonts to be used system wide in a number of predefined categories.

I also provide methods to interactively choose these fonts, to reset all fonts to some default value, to set larger demo fonts, to construct a font menu and to manipulate the current settings as an external specification.

The font categories that I know of are: ButtonFont, CodeFont, HaloFont, ListFont, MenuFont and WindowTitleFont

Examples:
- I open a dialog to choose the font for buttons with:
StandardFonts chooseButtonFont.

- To change base size for all fonts (when you want big fonts for a screencast or a presentation):
StandardFonts setDemoFonts.

- Then restore default fonts with:
StandardFonts restoreDefaultFonts.!

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 09:36'!
balloonFont
	^BalloonMorph balloonFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 09:36'!
balloonFont: aFont
	BalloonMorph balloonFont: aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/28/2009 21:47'!
buttonFont
	^ ButtonFont ifNil: [ButtonFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:07'!
buttonFont: aFont
	 ButtonFont := aFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:50'!
chooseBalloonFont
	self chooseStandardFont: #balloonFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:51'!
chooseButtonFont
	self chooseStandardFont: #buttonFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:51'!
chooseCodeFont
	self chooseStandardFont: #codeFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:48'!
chooseDefaultFont
	self chooseStandardFont: #defaultFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/18/2009 13:58'!
chooseFontWithTitle: aTitle for: aReceiver setSelector: aSetSelector getSelector: aGetSelector
	FontChooser
		openWithWindowTitle: aTitle
		for: aReceiver
		setSelector: aSetSelector
		getSelector: aGetSelector! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:49'!
chooseHaloFont
	self chooseStandardFont: #haloFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:49'!
chooseListFont
	self chooseStandardFont: #listFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:50'!
chooseMenuFont
	self chooseStandardFont: #menuFont! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/18/2009 13:59'!
chooseStandardFont: aSelector 
	self
		chooseFontWithTitle: (self systemFontLabelOf: aSelector)	, '...'
		for: self
		setSelector: (aSelector , ':') asSymbol
		getSelector: aSelector! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:50'!
chooseWindowTitleFont
	self chooseStandardFont: #windowTitleFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:22'!
codeFont
	^ CodeFont ifNil: [CodeFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:22'!
codeFont: aFont
	 CodeFont := aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 19:15'!
defaultFont
	^ TextStyle defaultFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 9/13/2011 21:32'!
defaultFont: aFont
	"Establish the default text font and style"

	| aStyle newDefaultStyle |
	aFont ifNil: [^ StandardFonts setFontsFromSpec: {{#defaultFont:. StrikeFont defaultFontKey. StrikeFont defaultFontSize}}].
	aStyle := aFont textStyle ifNil: [^ self].
	newDefaultStyle := aStyle copy.
	newDefaultStyle defaultFontIndex: (aStyle fontIndexOf: aFont).
	TextSharedInformation at: #DefaultTextStyle put: newDefaultStyle.
! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/9/2009 09:24'!
getCurrentFontsSpec
	"self getCurrentFontsSpec"
	^ self systemFontAspects collect: [:aspect | | font |
		font := self perform: aspect.
		Array with: (aspect, ':') asSymbol with: font familyName with: font pointSize]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:19'!
haloFont
	^ HaloFont ifNil: [HaloFont := StandardFonts defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:20'!
haloFont: aFont
	HaloFont := aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:24'!
listFont
	^ ListFont ifNil: [ListFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 10:24'!
listFont: aFont
	 ListFont := aFont! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/29/2009 11:55'!
menuFont
	^ MenuFont ifNil: [MenuFont := self defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'StephaneDucasse 7/2/2010 19:01'!
menuFont: aFont
	aFont ~= MenuFont 
		ifTrue: [ MenuFont := aFont ].! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/19/2009 17:56'!
printStandardSystemFonts
	"self printStandardSystemFonts"

	| string |
	string := String streamContents: [ :s |

	StandardFonts systemFontAspects do: [:selector |
		| font |
		font := StandardFonts perform: selector.
		s
			nextPutAll: selector; space;
			nextPutAll: font familyName; space;
			nextPutAll: (AbstractFont emphasisStringFor: font emphasis);
			nextPutAll: ' points: ';
			print: font pointSize;
			nextPutAll: ' height: ';
			print: font height;
			cr
		]].

	(StringHolder new)
		contents: string;
		openLabel: 'Current system font settings' translated.
! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 9/13/2011 21:36'!
restoreDefaultFonts
	"Since this is called from menus, we can take the opportunity to prompt for missing font styles."
	"
	self restoreDefaultFonts
	"
	| k |
	k := StrikeFont defaultFontKey.
	self setFontsFromSpec: {
			{#defaultFont:. k. 9 }. 
			{#codeFont:. k. 9 }.
			{#listFont:. k. 9 }. 
			{#menuFont:. k. 9 }. 
			{#windowTitleFont:. k. 12 }. 
			{#balloonFont:. k. 9 }. 
			{#haloFont:. k. 9 }.
			{#buttonFont:. k. 9 } }
! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/9/2009 08:46'!
setAllStandardFontsTo: aFont
	self systemFontAspects do: [:selector | StandardFonts perform: (selector, ':') asSymbol with: aFont].
	 
! !

!StandardFonts class methodsFor: 'menu' stamp: 'MarcusDenker 8/15/2010 13:11'!
setDefaultFonts: defaultFontsSpec
	"Since this is called from menus, we can take the opportunity to prompt for missing font styles."

	| fontNames map emphases |
	fontNames := defaultFontsSpec collect: [:array | array second].
	map := IdentityDictionary new.
	emphases := IdentityDictionary new.
	fontNames do: [:originalName | | decoded style response |
		decoded := TextStyle decodeStyleName: originalName.
		style := map at: originalName put: (TextStyle named: decoded second).
		emphases at: originalName put: decoded first.
		style ifNil: [map at: originalName put: (TextStyle default) ]].
	defaultFontsSpec do: [:triplet | self
		perform: triplet first
		with: (((map at: triplet second) fontOfPointSize: triplet third) emphasized: (emphases at: triplet second))]! !

!StandardFonts class methodsFor: 'choosing' stamp: 'AlainPlantec 12/19/2009 17:52'!
setDemoFonts
	"StandardFonts setDemoFonts"
	|size font codeFont titleFont|

	size := UIManager default request: 'Base font size?' initialAnswer: '14'.
	size isEmptyOrNil ifTrue: [^ self].
	size := size asInteger.
	(size isNil or: [size <= 0]) ifTrue: [^ self].
	font := LogicalFont familyName: 'DejaVu Sans' pointSize: size.
	codeFont := LogicalFont familyName: 'DejaVu Sans Mono' pointSize: size.
	titleFont := LogicalFont familyName: 'DejaVu Serif' pointSize: size.

	StandardFonts
		listFont: font;
		menuFont: font;
		codeFont: codeFont;
		buttonFont: font;
		defaultFont: font;
		windowTitleFont: titleFont.! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/9/2009 09:22'!
setFontsFromSpec: fontsSpec
	"set some fonts by interpreting fontsSpec.
	fontsSpec is a collection.
	Each fontsSpec element is an array with 3 elements: 
	(1) the font accessor (i.e. #codeFont:)
	(2) the font family (i.e. 'Arial')
	(3) the font point size.
	To see an example of fontsSpec, inspect the result of 
	StandardFonts getCurrentFontsSpec.
	"
	| fontNames map emphases |
	fontNames := fontsSpec
				collect: [:array | array second].
	map := IdentityDictionary new.
	emphases := IdentityDictionary new.
	fontNames
		do: [:originalName | 
			| decoded style response |
			decoded := TextStyle decodeStyleName: originalName.
			style := map
						at: originalName
						put: (TextStyle named: decoded second).
			emphases at: originalName put: decoded first.
			style
				ifNil: [map at: originalName put: TextStyle default]].
	fontsSpec
		do: [:triplet | self
				perform: triplet first
				with: (((map at: triplet second)
						fontOfPointSize: triplet third)
						emphasized: (emphases at: triplet second))]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/18/2009 13:49'!
systemFontAspects
	^ self systemFontAspectsAndLabels collect: [:assoc | assoc key]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/18/2009 13:53'!
systemFontAspectsAndLabels
	^ {#defaultFont -> 'Default font' translated. #listFont -> 'List font'  translated. #menuFont -> 'Menu font'  translated. #windowTitleFont -> 'Window title'  translated. #balloonFont -> 'Balloon help font'  translated. #codeFont -> 'Code font'  translated. #buttonFont -> 'Button font'  translated. #haloFont -> 'Halo font'  translated } ! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/18/2009 13:51'!
systemFontLabelOf: aSelector
	^ (self systemFontAspectsAndLabels detect: [:assoc | assoc key = aSelector] ifNone: [^ '']) value! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 12/19/2009 18:10'!
systemFontLabels
	^ self systemFontAspectsAndLabels collect: [:assoc | assoc value]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:21'!
windowTitleFont
	^ WindowTitleFont ifNil: [WindowTitleFont := StandardFonts defaultFont]! !

!StandardFonts class methodsFor: 'font accessing' stamp: 'AlainPlantec 11/30/2009 17:21'!
windowTitleFont: aFont
	WindowTitleFont := aFont! !

Object subclass: #TextStopConditions
	instanceVariableNames: 'stops crossedX endOfRun'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Text'!

!TextStopConditions commentStamp: 'nice 3/8/2010 14:31' prior: 0!
A TextStopConditions is a private helper class for text composition (See class CharacterScanner and subclasses).
It maps some characters controling the text layout (like carriage return, line feeds, space and tabulation) to some selector representing the action to be performed by the CharacterScanner.
By default, a TextStopConditions does not map any action to control character; instances must be properly initialized by sending #at:put: messages.

For example, inter-word spacing can be adjusted so as to obtain "justified" paragraphs. See implementors of #paddedSpace, #cr, #space, #columnBreak for example of special character actions.

A TextStopConditions also store two selectors for mapping the actions to be taken when:
- end of run is encountered;
- text overflows horizontal composition bounds.
These actions are by default #endOfRun and #crossedX (see implementors of these messages), but can enventually be changed using #endOfRun: and #crossedX:.

In a text (see class Text), the "runs" are used to store text style attributes, so an "end of run" event probably means some action in the textcomposer should be taken to change the font.

TextStopConditions current implementation can only map 256 character codes (from 1 to 256). It is the composer responsibility to encode the character before sending #at:. Presumably, the composer will use the character codePoint + 1 (see implementors of #codePoint).
If this is not sufficient, then this class could be changed to use a Dictionary or a LargeTable

Historically, the EndOfRun and CrossedX were two TextConstant of value 257 and 258, which did occupy corresponding slots in the stops array. Since these are valid character codePoint, this usage has been deprecated. However, because any error in text composition would have catastrophic consequences (unresponsive user interface), backward compatibility with obsolete historical code is still maintained internally which is why the stops array has a sze of 258.

Instance Variables
	crossedX:		<Symbol | nil>
	endOfRun:		<Symbol | nil>
	stops:		<Array of: Symbol | nil>

crossedX
	- selector to perform when the composed text overflows X composition bound

endOfRun
	- selector to perform at end of run

stops
	- an array mapping character code (codePoint + 1) to special actions, or nil if character is to be rendered normally
!

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:01'!
at: anInteger
	"Answer the special action associated with a character of code anInteger, or nil if none.
	The character code should be betxween 1 and 256 (presumably codePoint + 1).
	It can eventually be 257 for endOfRun action, or 258 for crossedX action for backward compatibility with historical squeak versions."
	
	^stops at: anInteger! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:02'!
at: anInteger put: aSymbolOrNil
	"Set the special action associated with a character of code anInteger, or nil if none.
	The character code should be betxween 1 and 256 (presumably codePoint + 1).
	It can eventually be 257 for endOfRun action, or 258 for crossedX action for backward compatibility with historical squeak versions."
	
	anInteger = 257
		ifTrue: [
			self endOfRun: aSymbolOrNil..
			^aSymbolOrNil].
	anInteger = 258
		ifTrue: [
			self crossedX: aSymbolOrNil..
			^aSymbolOrNil].
	^stops at: anInteger put: aSymbolOrNil! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:01'!
crossedX
	"Answer the special action to be performed when crossing composition bounds."
	
	^crossedX! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 12:00'!
crossedX: aSymbolOrNil
	crossedX := aSymbolOrNil.
	
	"Backward compatibility with historical EndOfRun TextConstant handling"
	stops size >= 258 ifTrue: [stops at: 258 put: crossedX]! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 14:01'!
endOfRun
	"Answer the special action to be performed et end of text."
	
	^endOfRun! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 12:00'!
endOfRun: aSymbolOrNil
	endOfRun := aSymbolOrNil.
	
	"Backward compatibility with historical EndOfRun TextConstant handling"
	stops size >= 257 ifTrue: [stops at: 257 put: endOfRun]! !

!TextStopConditions methodsFor: 'initialize-release' stamp: 'nice 3/8/2010 14:31'!
initialize
	"Initialize the default stop conditions."
	stops := Array new: 258.
	self endOfRun: #endOfRun.
	self crossedX: #crossedX.! !

!TextStopConditions methodsFor: 'copying' stamp: 'nice 4/17/2009 00:52'!
postCopy
	stops := stops copy! !

!TextStopConditions methodsFor: 'initialize-release' stamp: 'nice 2/28/2010 19:26'!
setStops: stopArray
	stops := stopArray! !

!TextStopConditions methodsFor: 'accessing' stamp: 'nice 3/8/2010 23:20'!
size
	"Return the highest code to which an action can be mapped.
	Do not take into account the two slots for endOfRun and crossedX, since they don't map any character code.
	Warning: this message is used for compatibility with obsolete code."
	
	^256! !
CharacterScanner initialize!
